#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <fcntl.h>	// O_RDONLY

#define VERSION_STR "0.1"
#define DELIM_CHAR ' '

#define NL "\n"

//==============
#if 0
void dprintf(const char *fmt, ...) {
    va_list ap;
    va_start(ap, fmt);
    vprintf(fmt, ap);
    va_end(ap);
}
#else
extern void dprintf(const char *fmt, ...) {}
#endif

//==============
int gen;
char out_buffer[4096];
char *cur_ptr = out_buffer;

extern inline void reset_buf() {
    cur_ptr = out_buffer;
}

extern inline void print_buf() {
    if(cur_ptr!=out_buffer) {
	write(1, out_buffer+1, cur_ptr-out_buffer-1);
	cur_ptr = out_buffer;
    }
}

void put(const char *s) {
    int sz = strlen(s);
    if(sz > (out_buffer+sizeof(out_buffer))-cur_ptr)
	sz = (out_buffer+sizeof(out_buffer))-cur_ptr;
    memcpy(cur_ptr, s, sz);
    cur_ptr += sz;
}

void put_c(char c) {
    if(cur_ptr < out_buffer+sizeof(out_buffer))
	*cur_ptr++ = c;
}

//==============
int readfile_z(char *buf, int sz, const char* fname) {
//==============
    int fd;
    fd=open(fname,O_RDONLY);
    if(fd<0) return 1;
    sz = read(fd,buf,sz-1);
    close(fd);
    if(sz<0) return 1;
    buf[sz]=0;
    return 0;
}

//==============
int rdval(const char* p, const char* key, unsigned long *vec, ...) {
//==============
    int strlen_key = strlen(key);
    do {
        while(isspace(*p) && *p!='\0') p++;
        if(strncmp(p, key, strlen_key) != 0) goto nextline;

        va_list arg_ptr;
        int indexline = 1;
        int indexnext;

        p += strlen_key;

        va_start(arg_ptr, vec);
        indexnext = va_arg(arg_ptr, int);
        while(1) {
    	    while(isspace(*p) && *p!='\n' && *p!='\0') p++;
	    if(*p=='\n' || *p=='\0') break;

            if(indexline == indexnext) {
                *vec++ = strtoul(p, NULL, 10);
                indexnext = va_arg(arg_ptr, int);
            }
    	    while(!isspace(*p) && *p!='\n' && *p!='\0') p++;
            indexline++;
        }
        va_end(arg_ptr);

        return 0;

    nextline:
        while(*p!='\n' && *p!='\0') p++;

    } while(*p!='\0');

    return 1;
}

//==============
#define SCALE_BAD "????"

//==============
void scale(unsigned long ul) {
//==============
    char buf[5];
    int index = 0;
    ul *= 10;
    while(ul >= 10000) {
        ul /= 1024;
        index++;
    }

    /* We'll use NNN format if: */
    if( !index /* it's not kilo/mega/giga - no point in printing '5.0 bytes' */
    || ul>=100 /* or if number is >= 10.0 - wont fit in N.N */
    ) {
	//sprintf(buf
        //    , "%3d%c"
        //    , ul / 10
        //    , " kMGTEP"[index]
	//);
	//put(buf);
	if( (buf[0]= " 123456789"[ul/1000]) == ' ')
	    buf[1] = " 123456789"[ul/100%10];
	else
	    buf[1] = "0123456789"[ul/100%10];
	buf[2] = "0123456789"[ul/10%10];
    } else { /* Do N.N format */
	//sprintf(buf
        //    , "%d.%d%c"
        //    , ul / 10
        //    , ul % 10
    	//    , " kMGTEP"[index]
	//);
	buf[0] = "0123456789"[ul/10];
	buf[1] = '.';
	buf[2] = "0123456789"[ul%10];
    }
    buf[3] = " kMGTEP"[index];
    buf[4] = 0;
    put(buf);
}

//==============
typedef struct s_stat {
    struct s_stat *next;
    void (*collect)(struct s_stat *s);
} s_stat;

//==============
typedef struct cpu_stat {
    struct cpu_stat *next;
    void (*collect)(struct cpu_stat *s);
    long old[4];
    //long total;
    int bar_sz;
    char *bar;
} cpu_stat;

//==============
void collect_cpu(cpu_stat *s) {
//==============
    char filebuf[4096];
    long cur[4];
    long frac[4];
    long all = 0;
    int norm_all = 0;
    int bar_sz = s->bar_sz;
    char *bar = s->bar;
    int i;

    put("cpu [");

    if(readfile_z(filebuf, sizeof(filebuf), "/proc/stat")) filebuf[0]=0;
    if(rdval(filebuf, "cpu", cur, 1, 2, 3, 4)) {
	memset(bar,'X',bar_sz);
	goto put_exit;
    }
    
    //dprintf("cur1:");
    for(i=0; i<4; i++) {
	long work;
	long old = s->old[i];
	if(cur[i] < old) old = cur[i];	//sanitize
        work = cur[i] - old;
        s->old[i] = cur[i];
        cur[i] = work;
        all += work;
	//dprintf(" %lu",cur[i]);
    }
    //dprintf(" all %lu\n",all);
    //s->total = all;

    if(all) {
	//dprintf("cur2:");
	for(i=0; i<4; i++) {
	    cur[i] = bar_sz*cur[i] / all;
	    frac[i] = bar_sz*cur[i] % all;
	    //dprintf(" %lu/%lu",cur[i],frac[i]);
	    norm_all += cur[i];
	}
	//dprintf(" norm_all %lu\n",norm_all);
    
	while(norm_all<bar_sz) {
	    int max=frac[0]; int pos=0;
	    //for(i=1; i<4; i++) if(frac[i]>=max) max=frac[i], pos=i;
	    if(frac[1]>=max) max=frac[1], pos=1;
	    if(frac[2]>=max) max=frac[2], pos=2;
	    if(frac[3]>=max) max=frac[3], pos=3;
	    frac[pos]=0;	//avoid bumping same value twice
	    cur[pos]++;
    	    //dprintf("bumped %i\n",pos);
	    norm_all++;
        }
    
	memset(bar,'.',bar_sz);
	memset(bar,'S',cur[2]); bar+=cur[2]; //sys
	memset(bar,'U',cur[0]); bar+=cur[0]; //usr
	memset(bar,'N',cur[1]); bar+=cur[1]; //nice
    } else {
	memset(bar,'?',bar_sz);
    }
put_exit:
    put(s->bar); put_c(']');
}

//==============
s_stat* init_cpu(const char *param) {
//==============
    int sz;
    cpu_stat *s = (cpu_stat*)malloc(sizeof(cpu_stat));
    s->next = 0;
    s->collect = &collect_cpu;

    sz = 10;
    if(isdigit(param[0])) {
	sz = strtoul(param, NULL, 0);
	if(sz<10) sz=10;
	if(sz>1000) sz=1000;
    }
    s->bar = (char*)malloc(sz+1);
    s->bar[sz]=0;
    s->bar_sz = sz;
    return (s_stat*)s;
}

//==============
char iffile[4096];
int ifgen;

typedef struct if_stat {
    struct if_stat *next;
    void (*collect)(struct if_stat *s);
    long old[4];
    const char *device;
    char *device_colon;
} if_stat;

//==============
void collect_if(if_stat *s) {
//==============
    long cur[4];
    int i;

    put(s->device);

    //dprintf("reading %s\n",s->device_colon);

    if(ifgen!=gen) {
	ifgen = gen;
	if(readfile_z(iffile, sizeof(iffile), "/proc/net/dev"))
	    iffile[0]=0;
    }

    if(rdval(iffile, s->device_colon, cur, 1, 3, 9, 11)) {
	put(" " SCALE_BAD " " SCALE_BAD);
	return;
    }

    //dprintf("cur1:");
    for(i=0; i<4; i++) {
	long work;
	long old = s->old[i];
	if(cur[i] < old) old = cur[i];	//sanitize
        work = cur[i] - old;
        s->old[i] = cur[i];
        cur[i] = work;
	//dprintf(" %lu",cur[i]);
    }
    //dprintf("\n");
    
    put_c(cur[1] ? '*' : ' ');
    scale(cur[0]);
    put_c(cur[3] ? '*' : ' ');
    scale(cur[2]);
}

//==============
s_stat* init_if(const char *device) {
//==============
    if_stat *s = (if_stat*)malloc(sizeof(if_stat));
    s->next = 0;
    s->collect = &collect_if;
    s->device = device;
    s->device_colon = (char*)malloc(strlen(device)+2);
    strcpy(s->device_colon,device);
    strcat(s->device_colon,":");
    return (s_stat*)s;
}

//==============
typedef struct mem_stat {
    struct mem_stat *next;
    void (*collect)(struct mem_stat *s);
} mem_stat;

//==============
void collect_mem() {
//==============
    unsigned long mem[3];
    char filebuf[4096];

    if(readfile_z(filebuf, sizeof(filebuf), "/proc/meminfo")) filebuf[0]=0;
    put("mem ");
    if(rdval(filebuf, "Mem:", mem, 2, 5, 6))
	put(SCALE_BAD);
    else
	scale(mem[0]-mem[1]-mem[2]);
}

//==============
s_stat* init_mem(const char *param) {
//==============
    mem_stat *s = (mem_stat*)malloc(sizeof(mem_stat));
    s->next = 0;
    s->collect = &collect_mem;
    return (s_stat*)s;
}

//==============
void run_once(s_stat *s) {
//==============
    gen++;
    while(s) {
        put_c(DELIM_CHAR);
        s->collect(s);
        s = s->next;
    }
}

//==============
int main(int argc, char* argv[]) {
//==============
    int sleeptime = 1000;
    s_stat *first = 0;
    s_stat *last = 0;
    char *final_str = "\n";
    int i;
    
    if(argc==1) {
	put(" "
	"Nanometer " VERSION_STR " allows you to monitor your system in real time" NL
	"Options:" NL
	"cNN     monitor CPU. NN - bar size in chars, default 10" NL
	"iIFACE  monitor network interface IFACE" NL
	"m       monitor memory" NL
	"dNNN    milliseconds between updates. Default 1000" NL
	"r       print <cr> instead of <lf> at EOL. Try it ;)" NL
	);
	print_buf();
	return 0;
    }

    for(i=1; i<argc; i++) {
	s_stat *s = 0;
	//dprintf("argv[%i]='%s'\n",i,argv[i]);
	switch(argv[i][0]) {
	case 'r':
	    final_str = "\r";
	    break;
	case 'd':
	    sleeptime = strtoul(argv[i]+1, NULL, 0);
	    break;
	case 'i':
	    s = init_if(argv[i]+1);
	    break;
	case 'c':
	    s = init_cpu(argv[i]+1);
	    break;
	case 'm':
	    s = init_mem(argv[i]+1);
	    break;
	default:
	    break;
	}
	if(s) {
	    if(!first) {
		first = (s_stat*)s;
	    } else {
		last->next = (s_stat*)s;
	    }
	    last = (s_stat*)s;
	}
    }
	
    run_once(first);
    reset_buf();
    while(1) {
	run_once(first);
	put(final_str);
	print_buf();
        usleep(sleeptime*1000);
    }

    return 0;
}
